@@ -813,9 +813,17 @@ CMSWindowsKeyState::fakeCtrlAltDel()
 		// current thread must be on that desktop to do the broadcast
 		// and we can't switch just any thread because some own windows
 		// or hooks.  so start a new thread to do the real work.
+		HANDLE hEvtSendSas = OpenEvent( EVENT_MODIFY_STATE, FALSE, "Global\\SendSAS" );
+		if ( hEvtSendSas ) {
+			LOG((CLOG_DEBUG "found the SendSAS event - signaling my launcher to simulate ctrl+alt+del"));
+			SetEvent( hEvtSendSas );
+			CloseHandle( hEvtSendSas );
+		}
+		else {
 		CThread cad(new CFunctionJob(&CMSWindowsKeyState::ctrlAltDelThread));
 		cad.wait();
 	}
+	}
 	else {
 		// simulate ctrl+alt+del
 		fakeKeyDown(kKeyDelete, KeyModifierControl | KeyModifierAlt,
@@ -1442,8 +1450,10 @@ CMSWindowsKeyState::getIDForKey(CKeyMap::KeyItem& item,
 		return id;
 
 	case 2:
-		// left over dead key in buffer;  oops.
-		return getIDForKey(item, button, virtualKey, keyState, hkl);
+		// left over dead key in buffer. this used to recurse,
+		// but apparently this causes a stack overflow, so just
+		// return no key instead.
+		return kKeyNone;
 	}
 }
 
