@@ -37,6 +37,14 @@
 #include "CArchMiscWindows.h"
 #include <string.h>
 #include <pbt.h>
+#include "GameDeviceTypes.h"
+
+#if GAME_DEVICE_SUPPORT
+#define WIN32_LEAN_AND_MEAN
+#include <Windows.h>
+#include "XInput.h"
+#include "XInputHook.h"
+#endif
 
 //
 // add backwards compatible multihead support (and suppress bogus warning).
@@ -73,6 +81,11 @@
 #define PBT_APMRESUMEAUTOMATIC	0x0012
 #endif
 
+#if GAME_DEVICE_SUPPORT
+typedef DWORD (WINAPI *XInputGetStateFunc)(DWORD, XINPUT_STATE*);
+typedef DWORD (WINAPI *XInputSetStateFunc)(DWORD, XINPUT_VIBRATION*);
+#endif
+
 //
 // CMSWindowsScreen
 //
@@ -80,8 +93,9 @@
 HINSTANCE				CMSWindowsScreen::s_instance = NULL;
 CMSWindowsScreen*		CMSWindowsScreen::s_screen   = NULL;
 
-CMSWindowsScreen::CMSWindowsScreen(bool isPrimary) :
+CMSWindowsScreen::CMSWindowsScreen(bool isPrimary, bool noHooks, bool gameDevice) :
 	m_isPrimary(isPrimary),
+	m_noHooks(noHooks),
 	m_is95Family(CArchMiscWindows::isWindows95Family()),
 	m_isOnScreen(m_isPrimary),
 	m_class(0),
@@ -106,6 +120,31 @@ CMSWindowsScreen::CMSWindowsScreen(bool isPrimary) :
 	m_keyState(NULL),
 	m_hasMouse(GetSystemMetrics(SM_MOUSEPRESENT) != 0),
 	m_showingMouse(false)
+#if GAME_DEVICE_SUPPORT
+	// @todo - move game support stuff to new class
+	, m_xInputPollThread(NULL)
+	, m_xInputTimingThread(NULL)
+	, m_xInputFeedbackThread(NULL)
+	, m_gameButtonsLast(0)
+	, m_gameLeftTriggerLast(0)
+	, m_gameRightTriggerLast(0)
+	, m_gameLeftStickXLast(0)
+	, m_gameLeftStickYLast(0)
+	, m_gameRightStickXLast(0)
+	, m_gameRightStickYLast(0)
+	, m_gameLastTimingSent(0)
+	, m_gameTimingWaiting(false)
+	, m_gameFakeLag(0)
+	, m_gamePollFreq(kGamePollFreqDefault)
+	, m_gamePollFreqAdjust(0)
+	, m_gameFakeLagMin(kGamePollFreqDefault)
+	, m_gameTimingStarted(false)
+	, m_gameTimingFirst(0)
+	, m_gameFakeLagLast(0)
+	, m_gameTimingCalibrated(false)
+	, m_gameDevice(gameDevice)
+	, m_xinputModule(NULL)
+#endif
 {
 	assert(s_instance != NULL);
 	assert(s_screen   == NULL);
@@ -116,7 +155,8 @@ CMSWindowsScreen::CMSWindowsScreen(bool isPrimary) :
 			m_hookLibrary = openHookLibrary("synrgyhk");
 		}
 		m_screensaver = new CMSWindowsScreenSaver();
-		m_desks       = new CMSWindowsDesks(m_isPrimary,
+		m_desks       = new CMSWindowsDesks(
+							m_isPrimary, m_noHooks,
 							m_hookLibrary, m_screensaver,
 							new TMethodJob<CMSWindowsScreen>(this,
 								&CMSWindowsScreen::updateKeysCB));
@@ -146,6 +186,34 @@ CMSWindowsScreen::CMSWindowsScreen(bool isPrimary) :
 
 	// install the platform event queue
 	EVENTQUEUE->adoptBuffer(new CMSWindowsEventQueueBuffer);
+
+#if GAME_DEVICE_SUPPORT
+	if (m_gameDevice)
+	{
+		m_xinputModule = LoadLibrary("xinput1_3.dll");
+		if (m_xinputModule == NULL)
+		{
+			throw XScreenXInputFailure("could not load xinput library");
+		}
+
+		if (m_isPrimary)
+		{
+			// only capture xinput on the server.
+			m_xInputPollThread = new CThread(new TMethodJob<CMSWindowsScreen>(
+				this, &CMSWindowsScreen::xInputPollThread));
+		}
+		else
+		{
+			// check for queued timing requests on client.
+			m_xInputTimingThread = new CThread(new TMethodJob<CMSWindowsScreen>(
+				this, &CMSWindowsScreen::xInputTimingThread));
+
+			// check for waiting feedback state on client.
+			m_xInputFeedbackThread = new CThread(new TMethodJob<CMSWindowsScreen>(
+				this, &CMSWindowsScreen::xInputFeedbackThread));
+		}
+	}
+#endif
 }
 
 CMSWindowsScreen::~CMSWindowsScreen()
@@ -489,7 +557,7 @@ void CMSWindowsScreen::saveMousePosition(SInt32 x, SInt32 y) {
 	m_xCursor = x;
 	m_yCursor = y;
 
-	LOG((CLOG_DEBUG2 "saved mouse position for next delta: %+d,%+d", x,y));
+	LOG((CLOG_DEBUG5 "saved mouse position for next delta: %+d,%+d", x,y));
 }
 
 UInt32
@@ -498,7 +566,9 @@ CMSWindowsScreen::registerHotKey(KeyID key, KeyModifierMask mask)
 	// only allow certain modifiers
 	if ((mask & ~(KeyModifierShift | KeyModifierControl |
 				  KeyModifierAlt   | KeyModifierSuper)) != 0) {
-		LOG((CLOG_WARN "could not map hotkey id=%04x mask=%04x", key, mask));
+		// this should be a warning, but this can confuse users,
+		// as this warning happens almost always.
+		LOG((CLOG_DEBUG "could not map hotkey id=%04x mask=%04x", key, mask));
 		return 0;
 	}
 
@@ -524,7 +594,9 @@ CMSWindowsScreen::registerHotKey(KeyID key, KeyModifierMask mask)
 	UINT vk = m_keyState->mapKeyToVirtualKey(key);
 	if (key != kKeyNone && vk == 0) {
 		// can't map key
-		LOG((CLOG_WARN "could not map hotkey id=%04x mask=%04x", key, mask));
+		// this should be a warning, but this can confuse users,
+		// as this warning happens almost always.
+		LOG((CLOG_DEBUG "could not map hotkey id=%04x mask=%04x", key, mask));
 		return 0;
 	}
 
@@ -653,7 +725,7 @@ CMSWindowsScreen::getCursorCenter(SInt32& x, SInt32& y) const
 }
 
 void
-CMSWindowsScreen::fakeMouseButton(ButtonID id, bool press) const
+CMSWindowsScreen::fakeMouseButton(ButtonID id, bool press)
 {
 	m_desks->fakeMouseButton(id, press);
 }
@@ -676,6 +748,167 @@ CMSWindowsScreen::fakeMouseWheel(SInt32 xDelta, SInt32 yDelta) const
 	m_desks->fakeMouseWheel(xDelta, yDelta);
 }
 
+void
+CMSWindowsScreen::fakeGameDeviceButtons(GameDeviceID id, GameDeviceButton buttons) const
+{
+#if GAME_DEVICE_SUPPORT
+	if (m_gameDevice)
+	{
+		LOG((CLOG_DEBUG "fake game device buttons id=%d buttons=%d", id, buttons));
+		SetXInputButtons(id, buttons);
+	}
+#endif
+}
+
+void
+CMSWindowsScreen::fakeGameDeviceSticks(GameDeviceID id, SInt16 x1, SInt16 y1, SInt16 x2, SInt16 y2) const
+{
+#if GAME_DEVICE_SUPPORT
+	if (m_gameDevice)
+	{
+		LOG((CLOG_DEBUG "fake game device sticks id=%d s1=%+d,%+d s2=%+d,%+d", id, x1, y1, x2, y2));
+		SetXInputSticks(id, x1, y1, x2, y2);
+	}
+#endif
+}
+
+void
+CMSWindowsScreen::fakeGameDeviceTriggers(GameDeviceID id, UInt8 t1, UInt8 t2) const
+{
+#if GAME_DEVICE_SUPPORT
+	if (m_gameDevice)
+	{
+		LOG((CLOG_DEBUG "fake game device triggers id=%d t1=%d t2=%d", id, t1, t2));
+		SetXInputTriggers(id, t1, t2);
+	}
+#endif
+}
+
+void
+CMSWindowsScreen::queueGameDeviceTimingReq() const
+{
+#if GAME_DEVICE_SUPPORT
+	if (m_gameDevice)
+	{
+		LOG((CLOG_DEBUG "queue game device timing request"));
+		QueueXInputTimingReq();
+	}
+#endif
+}
+
+void
+CMSWindowsScreen::gameDeviceTimingResp(UInt16 freq)
+{
+#if GAME_DEVICE_SUPPORT
+	if (!m_gameTimingStarted)
+	{
+		// record when timing started for calibration period.
+		m_gameTimingFirst = (UInt16)(ARCH->time() * 1000);
+		m_gameTimingStarted = true;
+	}
+
+	m_gameTimingWaiting = false;
+	m_gameFakeLagLast = m_gameFakeLag;
+	m_gameFakeLag = (UInt16)((ARCH->time() - m_gameLastTimingSent) * 1000);
+	m_gameFakeLagRecord.push_back(m_gameFakeLag);
+
+	if (m_gameFakeLag < m_gameFakeLagMin)
+	{
+		// record the lowest value so that the poll frequency
+		// is adjusted to track.
+		m_gameFakeLagMin = m_gameFakeLag;
+	}
+	else if (m_gameFakeLag > (m_gameFakeLagLast * 2))
+	{
+		// if fake lag has increased significantly since the last
+		// timing, then we must have reached the safe minimum.
+		m_gameFakeLagMin = m_gameFakeLagLast;
+	}
+
+	// only change poll frequency if it's a sensible value.
+	if (freq > kGamePollFreqMin && freq < kGamePollFreqMax)
+	{
+		m_gamePollFreq = freq;
+	}
+
+	UInt16 timeSinceStart = ((UInt16)(ARCH->time() * 1000) - m_gameTimingFirst);
+	if (!m_gameTimingCalibrated && (timeSinceStart < kGameCalibrationPeriod))
+	{
+		// during the calibration period, increase polling speed
+		// to try and find the lowest lag value.
+		m_gamePollFreqAdjust = 1;
+		LOG((CLOG_DEBUG2 "calibrating game device poll frequency, start=%d, now=%d, since=%d",
+			m_gameTimingFirst, (int)(ARCH->time() * 1000), timeSinceStart));
+	}
+	else
+	{
+		// @bug - calibration seems to re-occur after a period of time,
+		// though, this would actually be a nice feature (but could be
+		// a bit risky -- could cause poor game play)... setting this
+		// stops calibration from happening again.
+		m_gameTimingCalibrated = true;
+
+		// only adjust poll frequency if outside desired limits.
+		m_gamePollFreqAdjust = 0;
+		if (m_gameFakeLag > m_gameFakeLagMin * 3)
+		{
+			m_gamePollFreqAdjust = 1;
+		}
+		else if (m_gameFakeLag < m_gameFakeLagMin)
+		{
+			m_gamePollFreqAdjust = -1;
+		}
+	}
+
+	LOG((CLOG_DEBUG3 "game device timing, lag=%dms, freq=%dms, adjust=%dms, min=%dms",
+		m_gameFakeLag, m_gamePollFreq, m_gamePollFreqAdjust, m_gameFakeLagMin));
+
+	if (m_gameFakeLagRecord.size() >= kGameLagRecordMax)
+	{
+		UInt16 v, min = 65535, max = 0, total = 0;
+		std::vector<UInt16>::iterator it;
+		for (it = m_gameFakeLagRecord.begin(); it < m_gameFakeLagRecord.end(); ++it)
+		{
+			v = *it;
+			if (v < min)
+			{
+				min = v;
+			}
+			if (v > max)
+			{
+				max = v;
+			}
+			total += v;
+		}
+
+		LOG((CLOG_INFO "game device timing, min=%dms, max=%dms, avg=%dms",
+			min, max, (UInt16)(total / m_gameFakeLagRecord.size())));
+		m_gameFakeLagRecord.clear();
+	}
+#endif
+}
+
+void
+CMSWindowsScreen::gameDeviceFeedback(GameDeviceID id, UInt16 m1, UInt16 m2)
+{
+#if GAME_DEVICE_SUPPORT
+
+	XInputSetStateFunc xInputSetStateFunc =
+		(XInputSetStateFunc)GetProcAddress(m_xinputModule, "XInputSetState");
+
+	if (xInputSetStateFunc == NULL)
+	{
+		throw XScreenXInputFailure("could not get function address: XInputSetState");
+	}
+
+	XINPUT_VIBRATION vibration;
+	ZeroMemory(&vibration, sizeof(XINPUT_VIBRATION));
+	vibration.wLeftMotorSpeed = m1;
+	vibration.wRightMotorSpeed = m2;
+	xInputSetStateFunc(id, &vibration);
+#endif
+}
+
 void
 CMSWindowsScreen::updateKeys()
 {
@@ -735,6 +968,7 @@ CMSWindowsScreen::createBlankCursor() const
 	// create a transparent cursor
 	int cw = GetSystemMetrics(SM_CXCURSOR);
 	int ch = GetSystemMetrics(SM_CYCURSOR);
+
 	UInt8* cursorAND = new UInt8[ch * ((cw + 31) >> 2)];
 	UInt8* cursorXOR = new UInt8[ch * ((cw + 31) >> 2)];
 	memset(cursorAND, 0xff, ch * ((cw + 31) >> 2));
@@ -888,7 +1122,7 @@ bool
 CMSWindowsScreen::onPreDispatchPrimary(HWND,
 				UINT message, WPARAM wParam, LPARAM lParam)
 {
-	LOG((CLOG_DEBUG2 "handling pre-dispatch primary"));
+	LOG((CLOG_DEBUG5 "handling pre-dispatch primary"));
 
 	// handle event
 	switch (message) {
@@ -1270,8 +1504,8 @@ CMSWindowsScreen::onMouseMove(SInt32 mx, SInt32 my)
 	SInt32 x = mx - m_xCursor;
 	SInt32 y = my - m_yCursor;
 
-	LOG((CLOG_DEBUG2
-		"handling mouse move; delta motion calc: %+d=(%+d - %+d),%+d=(%+d - %+d)",
+	LOG((CLOG_DEBUG3
+		"mouse move - motion delta: %+d=(%+d - %+d),%+d=(%+d - %+d)",
 		x, mx, m_xCursor, y, my, m_yCursor));
 
 	// ignore if the mouse didn't move or if message posted prior
@@ -1296,7 +1530,7 @@ CMSWindowsScreen::onMouseMove(SInt32 mx, SInt32 my)
 		// center on the server screen. if we don't do this, then the mouse 
 		// will always try to return to the original entry point on the 
 		// secondary screen.
-		LOG((CLOG_DEBUG2 "warping server cursor to center: %+d,%+d", m_xCenter, m_yCenter));
+		LOG((CLOG_DEBUG5 "warping server cursor to center: %+d,%+d", m_xCenter, m_yCenter));
 		warpCursorNoFlush(m_xCenter, m_yCenter);
 		
 		// examine the motion.  if it's about the distance
@@ -1748,6 +1982,157 @@ CMSWindowsScreen::wndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
 	return result;
 }
 
+#if GAME_DEVICE_SUPPORT
+
+void
+CMSWindowsScreen::xInputPollThread(void*)
+{
+	LOG((CLOG_DEBUG "xinput poll thread started"));
+
+	XInputGetStateFunc xInputGetStateFunc = 
+		(XInputGetStateFunc)GetProcAddress(m_xinputModule, "XInputGetState");
+
+	if (xInputGetStateFunc == NULL)
+	{
+		throw XScreenXInputFailure("could not get function address: XInputGetState");
+	}
+
+	int index = 0;
+	XINPUT_STATE state;
+	bool end = false;
+	while (!end)
+	{
+		ZeroMemory(&state, sizeof(XINPUT_STATE));
+		DWORD result = xInputGetStateFunc(index, &state);
+
+		// timeout the timing request after 10 seconds
+		if (m_gameTimingWaiting && (ARCH->time() - m_gameLastTimingSent > 2))
+		{
+			m_gameTimingWaiting = false;
+			LOG((CLOG_DEBUG "game device timing request timed out"));
+		}
+		
+		// xinput controller is connected
+		if (result == ERROR_SUCCESS)
+		{
+			// @todo game device state class
+			bool buttonsChanged = state.Gamepad.wButtons != m_gameButtonsLast;
+			bool leftTriggerChanged = state.Gamepad.bLeftTrigger != m_gameLeftTriggerLast;
+			bool rightTriggerChanged = state.Gamepad.bRightTrigger != m_gameRightTriggerLast;
+			bool leftStickXChanged = state.Gamepad.sThumbLX != m_gameLeftStickXLast;
+			bool leftStickYChanged = state.Gamepad.sThumbLY != m_gameLeftStickYLast;
+			bool rightStickXChanged = state.Gamepad.sThumbRX != m_gameRightStickXLast;
+			bool rightStickYChanged = state.Gamepad.sThumbRY != m_gameRightStickYLast;
+
+			m_gameButtonsLast = state.Gamepad.wButtons;
+			m_gameLeftTriggerLast = state.Gamepad.bLeftTrigger;
+			m_gameRightTriggerLast = state.Gamepad.bRightTrigger;
+			m_gameLeftStickXLast = state.Gamepad.sThumbLX;
+			m_gameLeftStickYLast = state.Gamepad.sThumbLY;
+			m_gameRightStickXLast = state.Gamepad.sThumbRX;
+			m_gameRightStickYLast = state.Gamepad.sThumbRY;
+
+			bool eventSent = false;
+
+			if (buttonsChanged)
+			{
+				LOG((CLOG_DEBUG "xinput buttons changed"));
+
+				// xinput buttons convert exactly to synergy buttons
+				sendEvent(getGameDeviceButtonsEvent(),
+					new CGameDeviceButtonInfo(index, state.Gamepad.wButtons));
+
+				eventSent = true;
+			}
+
+			if (leftStickXChanged || leftStickYChanged || rightStickXChanged || rightStickYChanged)
+			{
+				LOG((CLOG_DEBUG "xinput sticks changed"));
+
+				sendEvent(getGameDeviceSticksEvent(),
+					new CGameDeviceStickInfo(
+					index,
+					m_gameLeftStickXLast, m_gameLeftStickYLast,
+					m_gameRightStickXLast, m_gameRightStickYLast));
+
+				eventSent = true;
+			}
+
+			if (leftTriggerChanged || rightTriggerChanged)
+			{
+				LOG((CLOG_DEBUG "xinput triggers changed"));
+
+				// @todo seems wrong re-using x/y for a single value...
+				sendEvent(getGameDeviceTriggersEvent(),
+					new CGameDeviceTriggerInfo(
+						index,
+						state.Gamepad.bLeftTrigger,
+						state.Gamepad.bRightTrigger));
+
+				eventSent = true;
+			}
+
+			if (/*eventSent && */!m_gameTimingWaiting && (ARCH->time() - m_gameLastTimingSent > .5))
+			{
+				sendEvent(getGameDeviceTimingReqEvent(), NULL);
+				m_gameLastTimingSent = ARCH->time();
+				m_gameTimingWaiting = true;
+
+				LOG((CLOG_DEBUG "game device timing request at %.4f", m_gameLastTimingSent));
+			}
+		}
+
+		UInt16 sleep = m_gamePollFreq + m_gamePollFreqAdjust;
+		LOG((CLOG_DEBUG5 "xinput poll sleeping for %dms", sleep));
+		Sleep(sleep);
+	}
+}
+
+void
+CMSWindowsScreen::xInputTimingThread(void*)
+{
+	LOG((CLOG_DEBUG "xinput timing thread started"));
+
+	bool end = false;
+	while (!end)
+	{
+		// if timing request was queued, a timing response is queued
+		// when the xinput status is faked; if it was faked, go tell
+		// the server when this happened.
+		if (DequeueXInputTimingResp())
+		{
+			LOG((CLOG_DEBUG "dequeued game device timing response"));
+			sendEvent(getGameDeviceTimingRespEvent(),
+				new CGameDeviceTimingRespInfo(GetXInputFakeFreqMillis()));
+		}
+
+		// give the cpu a break.
+		Sleep(1);
+	}
+}
+
+void
+CMSWindowsScreen::xInputFeedbackThread(void*)
+{
+	LOG((CLOG_DEBUG "xinput feedback thread started"));
+
+	int index = 0;
+	bool end = false;
+	while (!end)
+	{
+		WORD leftMotor, rightMotor;
+		if (DequeueXInputFeedback(&leftMotor, &rightMotor))
+		{
+			LOG((CLOG_DEBUG "dequeued game device feedback"));
+			sendEvent(getGameDeviceFeedbackEvent(),
+				new CGameDeviceFeedbackInfo(index, leftMotor, rightMotor));
+		}
+
+		Sleep(50);
+	}
+}
+
+#endif
 
 //
 // CMSWindowsScreen::CHotKeyItem
